﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;
using wclBluetooth;
using wclCommon;

namespace GattClient
{
    internal class Client
    {
        private WorkingThread FThread;
        private Int64 FAddress;
        private wclBluetoothRadio FRadio;
        private ManualResetEvent FEvent;
        private wclGattClient FClient;

        private void Proc()
        {
            // Create GATT client and assign event handlers.
            FClient = new wclGattClient();
            FClient.OnConnect += Client_OnConnect;
            FClient.OnDisconnect += Client_OnDisconnect;
            FClient.OnCharacteristicChanged += Client_OnCharacteristicChanged;

            FThread.Trace("Connecting to client " + FAddress.ToString("X12"));
            FClient.Address = FAddress;
            Int32 Res = FClient.Connect(FRadio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                FThread.Trace("Connect to client " + FAddress.ToString("X12") + " failed.", Res);
                FEvent.Set();
            }

            // Wait for termination.
            IntPtr[] Handlers = { FEvent.SafeWaitHandle.DangerousGetHandle() };
            // Process APC until thread terminated.
            while (wclHelpers.WaitEx(Handlers, wclWinAPI.INFINITE) != wclWinAPI.WAIT_OBJECT_0)
                ;

            FClient.Disconnect();
            FThread.RemoveDevice(FAddress);

            FEvent.Close();
            FEvent = null;
            FClient = null;

            GC.Collect();
        }

        private void Client_OnCharacteristicChanged(object Sender, ushort Handle, byte[] Value)
        {
            FThread.Trace("Characteristic " + Handle.ToString("X4") + " value has been changed");
        }

        private void Client_OnDisconnect(object Sender, int Reason)
        {
            FThread.Trace("Client " + FAddress.ToString("X12") + " disconnected.", Reason);
            FEvent.Set();
        }

        private void Client_OnConnect(object Sender, int Error)
        {
            if (Error != wclErrors.WCL_E_SUCCESS)
            {
                FThread.Trace("Client connect failed.", Error);
                FEvent.Set();
                return;
            }

            // Read services.
            wclGattService[] Services;
            Int32 Res = FClient.ReadServices(wclGattOperationFlag.goNone, out Services);
            if (Res != wclErrors.WCL_E_SUCCESS)
                FThread.Trace("Failed to read services.", Res);
            else
            {
                if (Services == null)
                    FThread.Trace("  No services wee found.");
                else
                {
                    // Go through all services.
                    foreach (wclGattService Service in Services)
                    {
                        // Read characteristics.
                        wclGattCharacteristic[] Characteristics;
                        Res = FClient.ReadCharacteristics(Service, wclGattOperationFlag.goNone, out Characteristics);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            FThread.Trace("Read characteristics failed.", Res);
                        else
                        {
                            if (Characteristics == null)
                                FThread.Trace("No characteristics were found.");
                            else
                            {
                                // Go through characteristics.
                                foreach (wclGattCharacteristic Characteristic in Characteristics)
                                {
                                    // If characteristi is indicatable or notifiable - subscribe.
                                    if (Characteristic.IsIndicatable || Characteristic.IsNotifiable)
                                    {
                                        wclGattCharacteristic Char = Characteristic;
                                        if (Char.IsIndicatable && Char.IsNotifiable)
                                            Char.IsIndicatable = false;

                                        Int32 Result = FClient.Subscribe(Char);
                                        if (Result != wclErrors.WCL_E_SUCCESS)
                                            FThread.Trace("Subscribe failed.", Result);
                                        else
                                        {
                                            Result = FClient.WriteClientConfiguration(Char, true, wclGattOperationFlag.goNone);
                                            if (Res != wclErrors.WCL_E_SUCCESS)
                                                FThread.Trace("Write client configuration descriptor failed", Result);
                                            else
                                                FThread.Trace("SUBSCRIBED");
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            // If something went wrong - disconnect.
            if (Res != wclErrors.WCL_E_SUCCESS)
                FClient.Disconnect();
        }

        public Client(WorkingThread Thread, wclBluetoothRadio Radio, Int64 Address)
        {
            FThread = Thread;
            FRadio = Radio;
            FAddress = Address;
            FClient = null;

            FEvent = new ManualResetEvent(false);

            Thread thread = new Thread(Proc);
            thread.Start();
        }
    }
}
